#!/usr/bin/env python

""" Split datafame bam column into multiple columns """

import logging
import sys
import pandas as pd

from .group_create import group_create
from .group_mean import group_mean

def split_dataframe(df):
    """ Split datafame bam column into multiple columns """

    df['bam'] = df['bam'].astype(str)
    names     = df['bam'].tolist()
    names     = sorted(set(names))
    logging.debug('list of bams')
    logging.debug(names)

    df['pos'] = df['pos'].astype(int)
    positions = df['pos'].tolist()
    positions = sorted(set(positions))
    logging.debug('list of positions')
    logging.debug(positions)

    df['ref']  = df['ref'].astype(str)
    references = df['ref'].tolist()
    references = sorted(set(references))
    logging.debug('list of references')
    logging.debug(references)

    if len(references) > 1:
        logging.fatal('Bam files have more than one reference!')
        sys.exit(1)

    df_split = group_create(positions)

    logging.debug('The split dataframe is')
    logging.debug(df_split)

    for name in names:
        df_name = group_mean(df, name)
        df_split = pd.merge(df_split, df_name, left_on = 'group', right_on = 'group', how = 'outer')

    logging.debug('The split dataframe is')
    logging.debug(df_split)

    return df_split



#!/usr/bin/env python

""" Gets coverage for bam file """

import logging
import pandas as pd
import pysam

def genome_depth(meta):
    """ Takes a bam file file and gets coverage """

    # getting depth results from pysam
    df = pd.DataFrame([x.split('\t') for x in pysam.depth(meta['sorted_bam']).split('\n')]) # pylint: disable=E1101
    df.columns = ['ref', 'pos', 'cov']

    # ensuring that the type is correct
    df['pos'] = pd.to_numeric(df['pos'], errors='coerce')
    df['cov'] = pd.to_numeric(df['cov'], errors='coerce')
    df['bam'] = meta['file_name']
    df = df.dropna()

    logging.debug(f"pysam results for {meta['sorted_bam']}") # pylint: disable=W1203
    logging.debug(df)

    return df



#!/usr/bin/env python

""" Takes a bam file and 
    1. checks to see if bam file is paired or single end reads
    2. then reduces bam file to those reads in specified subregion
    2a. if paired, restricts reads to those that are correctly paired
    2b. if single, restricts reads to those that are within region """

import logging
import os
import pysam

def within(initial_bam, final_bam, subregion):
    """ reduces reads to those within subregion """

    logging.debug(f"Initial bam is {initial_bam}") # pylint: disable=W1203
    logging.debug(f"Final bam is {final_bam}") # pylint: disable=W1203
    logging.debug(f"Region is {subregion}" ) # pylint: disable=W1203

    if os.path.exists(initial_bam):
        single_check = int(pysam.view('-c', '-f',  '1', initial_bam))
        if single_check == 0 :
            pysam.view('-bh', '-o', final_bam, initial_bam, subregion, catch_stdout=False)
        else :
            pysam.view('-bh','-f2', '-o', final_bam, initial_bam, subregion, catch_stdout=False)



#!/usr/bin/env python

""" Takes a bam file and sorts and indexes it with pysam """

import os
import pysam

def prep(initial, final, threads):
    """ Sorts and indexes bam files listed in dic """

    if os.path.exists(initial):
        pysam.sort('-o', final, '-@', str(threads), initial)

    if os.path.exists(final):
        pysam.index(final)



#!/usr/bin/env python

""" Gets column names in order """

import logging
import pandas as pd

def column_names(bed):
    """ Takes a bed file, sorts it, and gets column names """

    df = pd.read_table(bed, header=None)
    df = df.sort_values([1, 2], ascending=[True, True])

    if len(df.columns) < 4:
        df[3] = df.index

    df[3] = df[3].astype(str)
    names = df[3].tolist()

    logging.debug("Names of amplicons :")
    logging.debug(names)

    return names



#!/usr/bin/env python

""" Creates groups """

import pandas as pd
import numpy as np

def group_create(positions):
    """ Creates groups for later averaging """

    df = pd.DataFrame([])
    df['pos'] = positions
    df['group'] = (df['pos'] / 500).apply(np.floor)
    df = df.drop('pos', axis=1)
    df = df.drop_duplicates()

    df['start'] = (df['group'] * 500).astype(int)
    df['end']   = (df['group'] * 500 + 499).astype(int)
    df['pos']   = df['start'].astype('str') + '-' + df['end'].astype('str')

    return df[['pos', 'group']]



#!/usr/bin/env python
# pylint: disable=logging-fstring-interpolation

""" Takes a bed file and turns the lines into strings """

import logging
import sys

def get_regions(meta, bed):
    """ Takes a bed file and turns the lines into strings """

    regions = []
    names   = []
    with open(bed, encoding="utf-8") as file:
        for line in file:
            ref   = str(line.split('\t')[0])
            start = str(int(line.split('\t')[1]))
            end   = str(int(line.split('\t')[2]))
            name  = str(line.split('\t')[3])
            regions.append(ref + ':' + start + ':' + end + ':' + name)
            names.append(name)

            bed = meta['tmp'] + name + '.bed'
            if int(start) <= 1:
                with open(bed, mode='wt', encoding="utf-8") as file:
                    file.write(ref + '\t' + str(int(end) + 1) + '\t5000000\n')
            else:
                with open(bed, mode='wt', encoding="utf-8") as file:
                    line1 = ref + '\t' + str('0') + '\t' + str(int(start) - 1) + '\n'
                    line2 = ref + '\t' + str(int(end) + 1) + '\t5000000\n'
                    logging.debug(f"line 1 is {line1.strip()} in {bed}")
                    logging.debug(f"line 2 is {line2.strip()} in {bed}")
                    file.write(line1 + line2)

    logging.debug('Regions extracted from bedfile:')
    logging.debug(regions)

    # making sure region names are unique
    if len(names) != len(sorted(set(names))):
        logging.critical("Names in bedfiles are not unique! Exiting.")
        sys.exit(1)

    return regions



#!/usr/bin/env python
# pylint: disable=logging-fstring-interpolation

""" Split processes by region in bedfile for concurrent """

import logging
import concurrent.futures
import pandas as pd

from .amplicon_depth import amplicon_depth
from .get_regions import get_regions
from .column_names import column_names

def amplicon_splitting(meta, args):
    """ Split processes by region in bedfile for concurrent """

    # getting region for parallel processing
    regions = get_regions(meta, args.bed)
    #regions = ['MN908947.3:54:385:1', 'MN908947.3:342:704:2', 'MN908947.3:664:1004:3', 'MN908947.3:965:1312:4', 'MN908947.3:1264:1623:5', 'MN908947.3:1595:1942:6' ]

    logging.info('Getting depth for amplicons')
    results = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=args.threads) as executor:
        tasks = []
        for bam in args.bam:
            for subregion in regions:
                future = executor.submit(amplicon_depth, meta[bam], subregion)
                tasks.append(future)

        completed_tasks, _ = concurrent.futures.wait(tasks,return_when=concurrent.futures.ALL_COMPLETED) # pylint: disable=C0301
        results = [task.result() for task in completed_tasks]

    # setting up the dataframe
    columns   = column_names(args.bed)
    df        = pd.DataFrame(columns= ['bam'] + columns)
    df['bam'] = meta['filenames']
    logging.debug('Initial empty dataframe:')
    logging.debug(df)

    # NOTE : Had to be done outside of concurrent
    for bam, name, cov in results :
        bamindex = df.index[df['bam'] == bam]
        df.loc[bamindex, name] = cov

    df = df.fillna(0)

    return df



#!/usr/bin/env python

""" Graphs the dataframe """

import matplotlib.pyplot as plt

def plotting_boxplot(df, d):
    """ graphs the dataframe """

    boxplot = df.boxplot(fontsize=5, rot=90, figsize=(15,8), grid=False)
    boxplot.plot()
    plt.title(d['title'])
    boxplot.set_ylabel(d['ylabel'])
    boxplot.set_xlabel(d['xlabel'])
    boxplot.figure.savefig(d['file'], bbox_inches='tight')
    plt.close()



#!/usr/bin/env python
# pylint: disable=logging-fstring-interpolation

""" Gets coverage for bam over subrange """

import logging
import os
import pysam

def get_coverage(bam, subrange):
    """ Gets coverage for bam over subrange """

    if os.path.exists(bam):
        # get the coverage of the region (finally!)
        logging.debug('Getting coverage for ' + bam + ' over ' + subrange)
        cov_line = pysam.coverage('--no-header', bam, '-r', subrange).strip()
        logging.debug(f"pysam coverage line : {cov_line}")
        cov = float(cov_line.split()[6])
    else:
        cov = 0.0

    logging.debug(f"The coverage for {bam} over {subrange} is {str(cov)}")

    return cov



#!/usr/bin/env python
# pylint: disable=logging-fstring-interpolation

""" Use pysam to get depth for amplicon region """

import logging
import os
import pysam
from .subregion import subregion
from .within import within
from .without import without
from .get_coverage import get_coverage

def amplicon_depth(meta, region):
    """ Use pysam to get depth for amplicon region """

    # getting subregion information and creating bedfile
    subrange, name  = subregion(region)

    after_reduction_bam      = meta['tmp'] + name + '.step1.' + meta['file_name']
    removing_outside_matches = meta['tmp'] + name + '.step2.' + meta['file_name']
    junk_bam                 = meta['tmp'] + name + '.step3.' + meta['file_name']
    final_bam                = meta['tmp'] + name + '.step4.' + meta['file_name']
    logging.debug('The filenames are going to be :')
    logging.debug(meta)

    # setting the default value
    cov = 0.0

    # reduce bam file to something smaller
    if os.path.exists(meta['sorted_bam']):
        logging.debug(f"Step 1. reducing bam for speed for {meta['sorted_bam']}")
        within(meta['sorted_bam'], after_reduction_bam, subrange)

    if os.path.exists(after_reduction_bam):
        pysam.index(after_reduction_bam)

        # remove all reads that fall outside of region of interest
        # warning : this is the slow part of the script
        logging.debug(f"Step 2. reducing bam for speed for {meta['sorted_bam']}")
        without(after_reduction_bam,
                removing_outside_matches,
                junk_bam,
                bed = meta['tmp'] + name + '.bed')

    if os.path.exists(removing_outside_matches):
        pysam.index(removing_outside_matches)

        # get only reads that are within subrange
        logging.debug(f"Step 3. final reduction for {meta['sorted_bam']}")
        within(removing_outside_matches, final_bam, subrange)

    if os.path.exists(final_bam):
        pysam.index(final_bam)
        cov = get_coverage(final_bam, subrange)

    logging.info(f"Amplicon bam file created for {meta['file_name']} over {subrange}")

    return [meta['file_name'], name, cov]



#!/usr/bin/env python

""" Convert region string in script to something useable """

import logging

def subregion(region):
    """ Creates string and bedfile """

    # create string
    ref   = region.split(':')[0]
    start = region.split(':')[1]
    end   = region.split(':')[2]
    reg   = ref + ':' + str(start) + '-' + str(end)

    name = region.split(':')[3]

    logging.debug(f"The region is {region}") # pylint: disable=W1203
    logging.debug(f"The subrange is {reg}") # pylint: disable=W1203
    logging.debug(f"The amplicon name is {name}") # pylint: disable=W1203

    return reg, name



#!/usr/bin/env python3
# pylint: disable=logging-fstring-interpolation

'''
Author: Erin Young

Description:

This script will find the coverage for each amplicon in a bedfile and then graph it.

EXAMPLE:
aci -b input.bam -d amplicon.bed -o out
'''

# I tried to keep dependencies down...
import argparse
import logging
import os
import sys
import tempfile
import pandas as pd

from aci.utils.amplicon_splitting   import amplicon_splitting
from aci.utils.genome_depth         import genome_depth
from aci.utils.plotting_amplicons   import plotting_amplicons
from aci.utils.plotting_depth       import plotting_depth
from aci.utils.prep                 import prep

# about 30 seconds per artic V3 primer on SRR13957125
# $ samtools coverage SRR13957125.sorted.bam
# #rname      startpos endpos numreads covbases coverage meandepth meanbaseq meanmapq
# MN908947.3  1        29903  1141595  29827    99.7458  5350.27   37.3      60
# 15000 - 16500

def main():
    """ Use pysam to get depth for amplicon region and general coverage """

    ##### ----- ----- ----- ----- ----- #####
    ##### Part 0. Setup                 #####
    ##### ----- ----- ----- ----- ----- #####

    version = '1.4.20240116'

    parser = argparse.ArgumentParser()
    parser.add_argument('-b', '--bam',
                        nargs = '+',
                        required = True,
                        type = str,
                        help = '(required) input bam file(s)')
    parser.add_argument('-d', '--bed',
                        required = True,
                        type = str,
                        help ='(required) amplicon bedfile')
    parser.add_argument('-o', '--out',
                        required = False,
                        type = str,
                        help = 'directory for results',
                        default = 'aci')
    parser.add_argument('-log', '--loglevel',
                        required = False,
                        type = str,
                        help = 'logging level',
                        default = 'INFO')
    parser.add_argument('-t', '--threads',
                        required = False,
                        type = int,
                        help = 'specifies number of threads to use',
                        default=4)
    parser.add_argument('-v', '--version',
                        help='print version and exit',
                        action = 'version',
                        version = version)
    args = parser.parse_args()

    logging.basicConfig(format='%(asctime)s - %(message)s',
        datefmt = '%y-%b-%d %H:%M:%S',
        level=args.loglevel.upper())

    if not os.path.exists(args.bed):
        logging.critical(f"bedfile {args.bed} does not exist. Exiting")
        sys.exit(2)

    if not os.path.exists(args.out):
        os.mkdir(args.out)

    logging.info(f"ACI version :\t\t{str(version)}")
    logging.info(f"Number of threads :\t{str(args.threads)}")
    logging.info(f"Final directory :\t\t{str(args.out)}")
    logging.info(f"Input bed file :\t\t{str(args.bed)}")
    logging.info(f"Input bam file(s) :\t{', '.join(args.bam)}")

    with tempfile.TemporaryDirectory(dir = args.out) as temp_dir:
        meta = {}
        filenames = []
        for bam in args.bam:
            meta[bam]                       = {}
            meta[bam]['initial_bam']        = bam
            meta[bam]['out']                = args.out
            meta['tmp']                     = temp_dir + '/'
            meta[bam]['tmp']                = temp_dir + '/'
            meta[bam]['file_name']          = os.path.basename(bam)
            meta[bam]['sorted_bam'] = meta[bam]['tmp'] + os.path.basename(bam)
            meta[bam]['sorted_bai'] = meta[bam]['sorted_bam'] + '.bai'
            filenames.append(meta[bam]['file_name'])

            logging.info(f"Sorting and indexing {meta[bam]['file_name']}")
            prep(meta[bam]['initial_bam'], meta[bam]['sorted_bam'], args.threads)
        logging.info('Finished sorting and indexing')
        meta['filenames'] = filenames

        logging.debug('the filenames for all the bam files are')
        logging.debug(filenames)

        ##### ----- ----- ----- ----- ----- #####
        ##### Part 1. Amplicon depths       #####
        ##### ----- ----- ----- ----- ----- #####

        df = amplicon_splitting(meta, args)

        plotting_amplicons(df, args.out)

        logging.info(f"Amplicon depth is saved in {args.out}/amplicon_depth.csv")
        logging.info(f"An boxplot of these depths is at {args.out}/amplicon_depth.png")

        # ##### ----- ----- ----- ----- ----- #####
        # ##### Part 2. Genome/bam depths     #####
        # ##### ----- ----- ----- ----- ----- #####

        df_pysam = pd.DataFrame([])

        # NOTE : Attempted with concurrent and this was just as fast

        for bam in args.bam:
            df_pysam_results = genome_depth(meta[bam])
            df_pysam = pd.concat([df_pysam, df_pysam_results], ignore_index=True)

        plotting_depth(df_pysam, args.out)

        logging.info(f"Genome depth is saved in {args.out}/genome_depth.csv")
        logging.info(f"An boxplot of these depths is at {args.out}/genome_depth.png")

        # ##### ----- ----- ----- ----- ----- #####
        # ##### Fin                           #####
        # ##### ----- ----- ----- ----- ----- #####

        logging.info('ACI is complete! (I hope all your primers are behaving as expected!)')

if __name__ == "__main__":
    main()



#!/usr/bin/env python

""" Takes a bam file and maps those within and without a region specified by a bedfile """

import pysam

def without(initial, unmatched, matched, bed):
    """ splits reads to those within (bam2) and without (bam1) subregion """

    pysam.view('-bh', initial, '-U', unmatched, '-o', matched, '-L', bed, catch_stdout=False) # pylint: disable=E1101



#!/usr/bin/env python

""" Graphs the dataframe """

import logging

from .plotting_boxplot import plotting_boxplot

def plotting_amplicons(df, out):
    """ graphs the dataframe """

    # writing results to a file
    df.to_csv(out + '/amplicon_depth.csv', index=False)

    # getting rid of column with all the bam names
    df = df.drop('bam', axis=1)
    df = df.astype(float)
    logging.debug(df)

    d = {
        'title': 'Primer Assessment',
        'ylabel': 'mean depth',
        'xlabel' : 'amplicon name',
        'file' : out + '/amplicon_depth.png'
    }

    plotting_boxplot(df,d)



#!/usr/bin/env python

""" Getting coverage ready for plotting """

import logging

from .split_dataframe   import split_dataframe
from .plotting_boxplot  import plotting_boxplot

def plotting_depth(df, out):
    """ Getting coverage ready for plotting """

    df = split_dataframe(df)

    # writing results to a file
    df.to_csv(out + '/overall_depth.csv', index=False)

    # restructuring for boxplot
    df['ungroup'] = df['group'] * 500
    df = df.drop('pos', axis=1)
    df = df.drop('group', axis=1)
    df = df.set_index('ungroup')
    df = df.transpose()

    logging.debug('Restructured coverage boxplot')
    logging.debug(df)

    d = {
        'title' : 'Overall coverage',
        'ylabel' : 'meandepth',
        'xlabel' : 'position',
        'file' : out + '/overall_depth.png'
    }

    plotting_boxplot(df,d)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='amplicon_coverage_inspector',
    version='1.4.20240116',
    author='Erin Young',
    author_email='eriny@utah.gov',
    url="https://github.com/erinyoung/ACI",
    description='Visualization of coverage for amplicon sequencing',
    long_description=long_description,
    long_description_content_type="text/markdown", 
    packages=find_packages(include=['aci', 'aci.*']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta ',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Bio-Informatics '
        ],
    keywords = [
        "bioinformatics",
        "amplicon",
        "coverage",
        "visualization",
    ],
    python_requires=">=3.8, <4",
    install_requires=[
        "pandas",
        "pysam",
        "matplotlib",
        "numpy",
    ],
    entry_points={
        'console_scripts': [
            'aci=aci.aci:main'
            ],
    },
)


#!/usr/bin/env python

""" Gets mean for each group """

import numpy as np

def group_mean(df, name):
    """ Gets mean for each group """
    df_name = df[df['bam'] == name].copy()
    df_name['group'] = (df_name['pos'] / 500).apply(np.floor)
    df_name = df_name[['group', 'cov']]
    df_name.columns = ['group', name]
    df_name = df_name.groupby(['group']).mean().reset_index()
    return df_name
